/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>

#include <hw/thermal.h>

#include "apps/apps.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/device-notifier.h"
#include "core/devices.h"

static int noti;

/* Dbus Interfaces */
#define THERMAL_PATH                   "/Org/Tizen/System/Pass/Monitor/Thermal"
#define THERMAL_INTERFACE              "org.tizen.system.pass.monitor.thermal"
#define THERMAL_METHOD_TEMP_CHANGED    "ChangedCooldownMode"

/* Action types */
#define RELEASE_ACTION                 "ReleaseAction"
#define WARNING_ACTION                 "WarningAction"
#define LIMIT_ACTION                   "LimitAction"
#define SHUTDOWN_ACTION                "ShutDownAction"

/* Dbus Interfaces (THERMAL_PATH, THERMAL_INTERFACE, and THERMAL_METHOD_TEMP_CHANGED)
   and Action Type Strings (RELEASE_ACTION, WARNING_ACTION, etc.)
   can be modified according to the thermal daemons (e.g., PASS and SIOP)

   #define THERMAL_PATH                 "/Org/Tizen/Trm/Siop"
   #define THERMAL_INTERFACE            "org.tizen.trm.siop"
   #define THERMAL_METHOD_TEMP_CHANGED	"ChangedCooldownMode" */

#define SIGNAL_OVERHEAT_TIME           "TimeUpdate"
#define OVERHEAT_CALLBACK_TIME		1 //seconds
#define OVERHEAT_POWEROFF_TIME		30

enum thermal_enum_type {
	NORMAL,
	WARNING,
	CRITICAL,
	DANGER
};

static int power_off(void)
{
	const struct device_ops *power;

	power = find_device("power");
	if (check_default(power))
		return -ENODEV;

	return power->execute("poweroff");
}

static void thermal_remove_noti(void)
{
	int ret = 0;

	if (noti) {
		ret = remove_notification("TempCooldownNotiOff", noti);
		if (ret < 0)
			_E("Fail to remove cooldown noti : %d", noti);
		else
			noti = 0;
	}
}

static void thermal_remove_popup(void)
{
	int temp;

	temp = launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "remove_cooldown_popups");
	if (temp < 0)
		_E("Fail to launch remove_cooldown_popups");
}

static void thermal_add_noti(void)
{
	thermal_remove_noti();
	if (!noti)
		noti = add_notification("TempCooldownNotiOn", NULL, NULL);
}

static void broadcast_thermal_state_legacy(const char *action_type)
{
	if (!action_type) {
		_E("Invalid action type");
		return;
	}

	_D("action_type = %s", action_type);
	dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_SYSNOTI,
			DEVICED_INTERFACE_SYSNOTI,
			"CoolDownChanged",
			g_variant_new("(s)", action_type));
}

static gboolean thermal_overheat_time_broadcast(void *data)
{
	int ret;
	static int time = OVERHEAT_POWEROFF_TIME;

	/* Transfer Time by Signal */
	ret = dbus_handle_broadcast_dbus_signal_var(POPUP_OVERHEAT_PATH, POPUP_OVERHEAT_INTERFACE,
			SIGNAL_OVERHEAT_TIME, g_variant_new("(i)", time));
	if (ret < 0)
		_E("Fail in updating overheat time");

	time -= 1;
	if (time < 0) {
		time = OVERHEAT_POWEROFF_TIME;
		ret = launch_system_app(APP_OVERHEAT, 2,
				APP_KEY_TYPE, "remove_overheat_popups");
		if (ret < 0)
			_E("Fail to remove overheat popups");

		ret = power_off();
		if (ret < 0)
			_E("Fail to power off");
		return G_SOURCE_REMOVE;
	} else
		return G_SOURCE_CONTINUE;
}

static void thermal_overheat_popup(void)
{
	int ret;
	guint timer;

	ret = launch_system_app(APP_OVERHEAT, 2,
			APP_KEY_TYPE, "overheat");

	if (ret < 0)
		_E("error to launch Overheat popup");
	else {
		timer = g_timeout_add_seconds(OVERHEAT_CALLBACK_TIME, thermal_overheat_time_broadcast, NULL);
		if (!timer)
			_E("Fail to set over temp timer");
	}
}

static void thermal_add_recovery_popup(void)
{
	int temp;

	temp = launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "cooled_down");
	if (temp < 0)
		_E("Fail to launch recovery popup");
}

static int change_templevel_to_enum(char *string)
{
	int temp = -1;

	if (!string) {
		_E("Invalid temp level");
		return temp;
	}

	if (!strncmp(string, RELEASE_ACTION, strlen(RELEASE_ACTION)))
		temp = NORMAL;
	else if (!strncmp(string, WARNING_ACTION, strlen(WARNING_ACTION)))
		temp = WARNING;
	else if (!strncmp(string, LIMIT_ACTION, strlen(LIMIT_ACTION)))
		temp = CRITICAL;
	else if (!strncmp(string, SHUTDOWN_ACTION, strlen(SHUTDOWN_ACTION)))
		temp = DANGER;

	return temp;
}

static void thermal_action(char *state)
{
	static int old_state = NORMAL;
	int new_state;
	char *action = NULL;

	new_state = change_templevel_to_enum(state);
	if (new_state < 0) {
		_E("Invalid thermal state : %d", new_state);
		goto out;
	}

	_D("old_state = %d, new_state = %d", old_state, new_state);
	if (old_state == DANGER)
		goto out;

	switch (new_state) {
	case NORMAL:
		if (old_state == WARNING || old_state == CRITICAL) {
			thermal_remove_noti();
			thermal_add_recovery_popup();
		}
			action = strdup(RELEASE_ACTION);
		break;

	case WARNING:
		thermal_remove_popup();
		thermal_add_noti();
		action = strdup(WARNING_ACTION);
		break;

	case CRITICAL:
		thermal_remove_popup();
		thermal_add_noti();
		action = strdup(LIMIT_ACTION);
		break;

	case DANGER:
		thermal_remove_popup();
		thermal_remove_noti();
		thermal_overheat_popup();
		action = strdup(SHUTDOWN_ACTION);
		break;

	default:
		_E("invalid state : %d", new_state);
		goto out;
	}

	old_state = new_state;
	broadcast_thermal_state_legacy(action);
	free(action);

out:
	_D("Last old_state = %d, new_state = %d", old_state, new_state);
}

static void temp_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	char *temp_str = NULL;

	g_variant_get(param, "(s)", &temp_str);

	_D("temp_str = %s", temp_str);
	thermal_action(temp_str);
	g_free(temp_str);
}

static guint sigid;

static int booting_done(void *data)
{
	static int done;
	int ret;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	_I("booting done");
	ret = subscribe_dbus_signal(NULL, THERMAL_PATH,
		THERMAL_INTERFACE,
		THERMAL_METHOD_TEMP_CHANGED,
		temp_signal_handler, NULL, NULL);
	if (ret <= 0)
		_E("fail to init dbus signal(%d)", ret);
	sigid = ret;

	return done;
}

static void thermal_init(void *data)
{
	int ret;

	ret = register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Fail to register booting done Notifier");
}

static void thermal_exit(void *data)
{
	int ret;

	ret = unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Fail to unregister booting done Notifier");

	if (sigid)
		unsubscribe_dbus_signal(NULL, sigid);
}

static int thermal_probe(void *data)
{
	/* device-manager-plugin-xxx is not used for the information of temperature.
	  instead, pass is used */

	_I("Thermal device structure load success");
	return 0;
}

static const struct device_ops thermal_ops = {
	.name     = "thermal",
	.probe    = thermal_probe,
	.init     = thermal_init,
	.exit     = thermal_exit,
};

DEVICE_OPS_REGISTER(&thermal_ops)
